/*
 *  Copyright 2007-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*
	This is the "persistent speculative while-while" kernel used in

	"Understanding the Efficiency of Ray Traversal on GPUs",
	Timo Aila and Samuli Laine,
	Proc. High-Performance Graphics 2009

	See README for information and comments.
*/

#include "HPG09_defs.h"

__device__	 int g_warpCounter;
__constant__ int c_rayCount = 0;

#define FETCH_GLOBAL(x, y) x[y]	
#define FETCH_TEXTURE(x, y) tex1Dfetch(tex##x, y)

texture<float4, 1> texrays;
texture<float4, 1> texnodeheaders;
texture<float4, 1> texinnernodesX;
texture<float4, 1> texinnernodesY;
texture<float4, 1> texinnernodesZ;
texture<float4, 1> textrianglesA;
texture<float4, 1> textrianglesB;
texture<float4, 1> textrianglesC;
texture<float,  1> textrianglesI;

__device__ void swap(int& a,int& b)	{int t=a; a=b; b=t;}

__global__ void trace(	RayResult* results, bool shadowRay, int* debugData1, int* debugData2, int* debugData3,
						float4* rays,
						float4* nodeheaders,
						float4* innernodesX,
						float4* innernodesY,
						float4* innernodesZ,
						float2* dummy,
						float4* trianglesA,
						float4* trianglesB,
						float4* trianglesC,
						float* trianglesI
						)
{
	//------------------------------------------------------
	// Shared and local mem arrays
	// - we steal 3 entries from the traversal stack for storing rarely used variables in order to reduce register pressure.
	//------------------------------------------------------

	__shared__ RayStruct shared[RAY_BLOCK_WIDTH*RAY_BLOCK_HEIGHT+1];

#define root 0
#define RAY_RESULT_POS		63
#define THREAD_IDX_X_POS	62
#define WARP_AUX_POS		61

	int traversalStack[RAY_RESULT_POS+1];

	//------------------------------------------------------
	// Live data in registers
	//------------------------------------------------------

	int traversalStackPtr,nodeAddr;
	int	leafAddr1,leafAddr2;
	float tTriangle;
	float origx,origy,origz;
	RayStruct* aux = shared + threadIdx.x + (RAY_BLOCK_WIDTH * threadIdx.y);

	traversalStack[THREAD_IDX_X_POS] = threadIdx.x;					// forced to local mem -> saves a register
	traversalStack[WARP_AUX_POS]	 = threadIdx.y;					// 

	//------------------------------------------------------
	// Persistent threads init
	//------------------------------------------------------

	__shared__ volatile int nextRayArray[RAY_BLOCK_HEIGHT];			// for broadcasting thread0 -> whole warp
	__shared__ volatile int rayCountArray[RAY_BLOCK_HEIGHT];		// for broadcasting thread0 -> whole warp

	{
		volatile int& localPoolNextRay = nextRayArray[threadIdx.y];
		volatile int& localPoolRayCount = rayCountArray[threadIdx.y];
		localPoolNextRay = 0;										// current ray index (in global buffer)
		localPoolRayCount = 0;										// how many rays left in local pool
	}

	//------------------------------------------------------
	// Start with an initial ray (tricks compiler to use less registers)
	// - Rays [0,LOAD_BALANCER_FIRST_WARP-1] assigned without atomics
	// - See README
	//------------------------------------------------------

	{
		int	firstRayIdx = RAY_BLOCK_WIDTH * (int)threadIdx.y + 
						  RAY_BLOCK_WIDTH * RAY_BLOCK_HEIGHT * ((int)blockIdx.x + RAY_GRID_WIDTH * (int)blockIdx.y);
		int rayidx = firstRayIdx + threadIdx.x;

		//-----------------------------------------------------
		// Fetch and init ray
		//-----------------------------------------------------

		float4 o = FETCH_GLOBAL(rays, rayidx);
		float4 d = FETCH_GLOBAL(rays, rayidx + RAY_BUFFER_SIZE);
		origx = o.x;
		origy = o.y;
		origz = o.z;
		float ooeps = 1/powf(2,-80);
		aux->idir_x  = d.x==0.f ? ooeps : 1/d.x;		// avoid div by 0
		aux->idir_y  = d.y==0.f ? ooeps : 1/d.y;
		aux->idir_z  = d.z==0.f ? ooeps : 1/d.z;
		RayResult* rayResult = &results[__float_as_int(d.w)];
		rayResult->tidx = -1;
		traversalStack[RAY_RESULT_POS] = (int)rayResult;

		//-----------------------------------------------------
		// Traversal init
		//-----------------------------------------------------

		traversalStack[0] = ENTRYPOINT_SENTINEL;		// init stack with a sentinel
		traversalStackPtr = 0;
		nodeAddr  = root;
		leafAddr1 = 0;
		leafAddr2 = 0;
		tTriangle = 1.f;
	}

  do
  {
	//------------------------------------------------------
	// TRAVERSE!
	//------------------------------------------------------

    int hitIndex = -1;
	int cnt = 0;
	while(nodeAddr!=ENTRYPOINT_SENTINEL)
	{
		//---------------------------------------------
		// Inner node
		//---------------------------------------------

		bool searchingLeaf = true;
		float ood_x  = origx * aux->idir_x;
		float ood_y  = origy * aux->idir_y;
		float ood_z  = origz * aux->idir_z;

		while(nodeAddr>=0 && nodeAddr!=ENTRYPOINT_SENTINEL)
		{
			N_AOS(float4 n0xy = FETCH_TEXTURE(innernodesX, nodeAddr*4+0);)	// node0: x0,x1,y0,y1
			N_AOS(float4 nz   = FETCH_TEXTURE(innernodesX, nodeAddr*4+2);)	// node0: z0,z1, node1: z0,z1
			N_AOS(float4 n1xy = FETCH_TEXTURE(innernodesX, nodeAddr*4+1);)	// node1: x0,x1,y0,y1
			N_AOS(float4 cnodes=FETCH_TEXTURE(innernodesX, nodeAddr*4+3);)
	
			N_SOA(float4 n0xy = FETCH_TEXTURE(innernodesX, nodeAddr);)		// node0: x0,x1,y0,y1
			N_SOA(float4 nz   = FETCH_TEXTURE(innernodesZ, nodeAddr);)		// node0: z0,z1, node1: z0,z1
			N_SOA(float4 n1xy = FETCH_TEXTURE(innernodesY, nodeAddr);)		// node1: x0,x1,y0,y1
			N_SOA(float4 cnodes=FETCH_TEXTURE(nodeheaders, nodeAddr);)

			float x0Child0 = n0xy.x * aux->idir_x - ood_x;
			float x1Child0 = n0xy.y * aux->idir_x - ood_x;
			float y0Child0 = n0xy.z * aux->idir_y - ood_y;
			float y1Child0 = n0xy.w * aux->idir_y - ood_y;
			float z0Child0 = nz.x   * aux->idir_z - ood_z;
			float z1Child0 = nz.y   * aux->idir_z - ood_z;
			float z0Child1 = nz.z   * aux->idir_z - ood_z;
			float z1Child1 = nz.w   * aux->idir_z - ood_z;
			float tminChild0 = fmaxf(0.f,      fminf(x0Child0, x1Child0));
			float tmaxChild0 = fminf(tTriangle,fmaxf(x0Child0, x1Child0));
			float tminChild1 = fmaxf(0.f,      fminf(z0Child1, z1Child1));
			float tmaxChild1 = fminf(tTriangle,fmaxf(z0Child1, z1Child1));
			tminChild0 = fmaxf(tminChild0, fminf(y0Child0, y1Child0));
			tmaxChild0 = fminf(tmaxChild0, fmaxf(y0Child0, y1Child0));
			tminChild0 = fmaxf(tminChild0, fminf(z0Child0, z1Child0));
			tmaxChild0 = fminf(tmaxChild0, fmaxf(z0Child0, z1Child0));
			float x0Child1 = n1xy.x * aux->idir_x - ood_x;
			float x1Child1 = n1xy.y * aux->idir_x - ood_x;
			float y0Child1 = n1xy.z * aux->idir_y - ood_y;
			float y1Child1 = n1xy.w * aux->idir_y - ood_y;
			tminChild1 = fmaxf(tminChild1, fminf(y0Child1, y1Child1));
			tmaxChild1 = fminf(tmaxChild1, fmaxf(y0Child1, y1Child1));
			tminChild1 = fmaxf(tminChild1, fminf(x0Child1, x1Child1));
			tmaxChild1 = fminf(tmaxChild1, fmaxf(x0Child1, x1Child1));

			//------------------------------------------------------
			// Traversal decision
			// - differs from "while-while" mainly because this just happened to produce better code here.
			//------------------------------------------------------

			bool traverseChild0 = (tmaxChild0 >= tminChild0);
			bool traverseChild1 = (tmaxChild1 >= tminChild1);

			nodeAddr		   = __float_as_int(cnodes.x);		// stored as int
			int nodeAddrChild1 = __float_as_int(cnodes.y);		// stored as int

			if(!traverseChild1)	{ nodeAddrChild1 = ENTRYPOINT_SENTINEL; }
			if(!traverseChild0)	{ nodeAddr = nodeAddrChild1; nodeAddrChild1 = ENTRYPOINT_SENTINEL; }

			if(nodeAddr == ENTRYPOINT_SENTINEL)
			{
				nodeAddr = traversalStack[traversalStackPtr];
				--traversalStackPtr;
			}
			else if(nodeAddrChild1 != ENTRYPOINT_SENTINEL)
			{
				if(tminChild1 < tminChild0)						// Ensure Child0 is near, Child1 is far
					swap(nodeAddr,nodeAddrChild1);				// Happens if child1 intersected AND closer
				++traversalStackPtr;
				traversalStack[traversalStackPtr] = nodeAddrChild1;		// push far
			}

			// postpone one leaf node
			if(nodeAddr<0 && leafAddr1>=0)
			{
				searchingLeaf = false;
				leafAddr1 = nodeAddr;
				nodeAddr = traversalStack[traversalStackPtr];
				--traversalStackPtr;
			}

			if(!__any(searchingLeaf))
				break;
		} 

		//---------------------------------------------------------
		// Ray-triangle intersection
		//---------------------------------------------------------

		while(leafAddr1<0)
		{
			N_AOS(float4 leaf=FETCH_TEXTURE(innernodesX, (-leafAddr1-1)*4+3);)
			N_SOA(float4 leaf=FETCH_TEXTURE(nodeheaders, (-leafAddr1-1));)
			int triAddr  = __float_as_int(leaf.x);				// stored as int
			int triAddr2 = __float_as_int(leaf.y);				// stored as int

			for( ;triAddr < triAddr2; triAddr++)
			{
				float dirx  = 1/aux->idir_x;
				float diry  = 1/aux->idir_y;
				float dirz  = 1/aux->idir_z;

				T_AOS(float4 v00 = FETCH_GLOBAL(trianglesA, triAddr*4+0);)
				T_AOS(float4 v11 = FETCH_GLOBAL(trianglesA, triAddr*4+1);)
				T_SOA(float4 v00 = FETCH_GLOBAL(trianglesA, triAddr);)
				T_SOA(float4 v11 = FETCH_GLOBAL(trianglesB, triAddr);)

				float Oz = v00.w - origx*v00.x - origy*v00.y - origz*v00.z;
				float invDz = 1.f / (dirx*v00.x + diry*v00.y + dirz*v00.z);
				float t = Oz * invDz;
				if (t>0.f && t < tTriangle)
				{
					float Ox = v11.w + origx*v11.x + origy*v11.y + origz*v11.z;
					float Dx = dirx*v11.x + diry*v11.y + dirz*v11.z;
					float u = Ox + t*Dx;
					if (u >= 0)
					{
						T_AOS(float4 v22 = FETCH_GLOBAL(trianglesA, triAddr*4+2);)
						T_SOA(float4 v22 = FETCH_GLOBAL(trianglesC, triAddr);)
						float Oy = v22.w + origx*v22.x + origy*v22.y + origz*v22.z;
						float Dy = dirx*v22.x + diry*v22.y + dirz*v22.z;
						float v = Oy + t*Dy;
						if (v >= 0 && (u+v) <= 1.f)
						{
							tTriangle = t;
							hitIndex = triAddr;
							if(shadowRay)
							{
								nodeAddr = ENTRYPOINT_SENTINEL;
								break;
							}
						}
					}
				}
			} // tri

			leafAddr1 = nodeAddr;
			if(nodeAddr<0)
			{
				nodeAddr = traversalStack[traversalStackPtr];
				--traversalStackPtr;
			}
		} // leaf
	}

	if(hitIndex!=-1)
	{
		RayResult* rayResult = (RayResult*)traversalStack[RAY_RESULT_POS];
		rayResult->tidx = __float_as_int(FETCH_GLOBAL(trianglesI, hitIndex));
		rayResult->t	= tTriangle;
	}

	//-----------------------------------------------------
	// Persistent threads
	//-----------------------------------------------------

	int tidx = traversalStack[THREAD_IDX_X_POS];
	int widx = traversalStack[WARP_AUX_POS];
	int popc = 32;
	int eidx = tidx;

	volatile int& localPoolRayCount = rayCountArray[widx];
	volatile int& localPoolNextRay = nextRayArray[widx];

	// Assign rays from global to local pool
	if(eidx==0 && localPoolRayCount<=0)	// local pool empty, eidx 0 fetches
	{
		localPoolNextRay = atomicAdd(&g_warpCounter, LOAD_BALANCER_BATCH_SIZE*32);
		localPoolRayCount = LOAD_BALANCER_BATCH_SIZE*32;
	}

	// Assign rays from local pool
	int rayidx = localPoolNextRay + eidx;
	if(rayidx >= c_rayCount)
		return;

	if(eidx==0)
	{
		localPoolNextRay += popc;
		localPoolRayCount -= popc;
	}

	//-----------------------------------------------------
	// Fetch and init ray
	//-----------------------------------------------------

	float4 o = FETCH_GLOBAL(rays, rayidx);
	float4 d = FETCH_GLOBAL(rays, rayidx + RAY_BUFFER_SIZE);
	origx = o.x;
	origy = o.y;
	origz = o.z;
	float ooeps = 1/powf(2,-80);
	aux->idir_x  = d.x==0.f ? ooeps : 1/d.x;	// avoid div by 0
	aux->idir_y  = d.y==0.f ? ooeps : 1/d.y;
	aux->idir_z  = d.z==0.f ? ooeps : 1/d.z;
	RayResult* rayResult = &results[__float_as_int(d.w)];
	rayResult->tidx = -1;
	traversalStack[RAY_RESULT_POS] = (int)rayResult;

	//-----------------------------------------------------
	// Traversal init
	//-----------------------------------------------------

	traversalStack[0] = ENTRYPOINT_SENTINEL;			// init stack
	traversalStackPtr = 0;
	nodeAddr  = root;
	leafAddr1 = 0;
	leafAddr2 = 0;
	tTriangle = 1.f;

  } while(1);
}
