/*
 *  Copyright 2007-2009 NVIDIA Corporation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

/*
	This is a simplified version of the "persistent while-while kernel" used in

	"Understanding the Efficiency of Ray Traversal on GPUs",
	Timo Aila and Samuli Laine,
	Proc. High-Performance Graphics 2009

	See README for information and comments.
*/

#include "HPG09_defs.h"

// Variables used by persistent threads
__device__	 int g_warpCounter;
__constant__ int c_rayCount = 0;

// While-while done by setting large numbers to these counters
// If-if follows from 1 & 1.
// It may be interesting to experiment with intermediate solutions as well (e.g. 8,8)
#define LOOP_I	100
#define LOOP_T	100

#define FETCH_GLOBAL(x, y) x[y]	
#define FETCH_TEXTURE(x, y) tex1Dfetch(tex##x, y)

texture<float4, 1> texrays;
texture<float4, 1> texnodeheaders;
texture<float4, 1> texinnernodesX;
texture<float4, 1> texinnernodesY;
texture<float4, 1> texinnernodesZ;
texture<float4, 1> textrianglesA;
texture<float4, 1> textrianglesB;
texture<float4, 1> textrianglesC;
texture<float,  1> textrianglesI;

__device__ void swap(int& a,int& b)	{int t=a; a=b; b=t;}

__global__ void trace(	RayResult* results, bool shadowRay, int* debugData1, int* debugData2, int* debugData3,
						float4* rays,
						float4* nodeheaders,
						float4* innernodesX,
						float4* innernodesY,
						float4* innernodesZ,
						float2* dummy,
						float4* trianglesA,
						float4* trianglesB,
						float4* trianglesC,
						float* trianglesI
						)
{
	//------------------------------------------------------
	// Shared and local mem arrays
	//------------------------------------------------------

	__shared__ RayStruct shared[RAY_BLOCK_WIDTH*RAY_BLOCK_HEIGHT+1];
	RayStruct* aux = shared + threadIdx.x + (RAY_BLOCK_WIDTH * threadIdx.y);

#define root 0
#define RAY_RESULT_POS		63		// steal one entry from stack (saves a register)
	int traversalStack[RAY_RESULT_POS+1];

	//------------------------------------------------------
	// Persistent threads init
	//------------------------------------------------------

	__shared__ volatile int nextRayArray[RAY_BLOCK_HEIGHT];			// for broadcasting thread0 -> whole warp
	__shared__ volatile int rayCountArray[RAY_BLOCK_HEIGHT];		// for broadcasting thread0 -> whole warp

	volatile int& localPoolNextRay = nextRayArray[threadIdx.y];
	volatile int& localPoolRayCount = rayCountArray[threadIdx.y];
	localPoolNextRay = 0;										// current ray index (in global buffer)
	localPoolRayCount = 0;										// how many rays left in local pool

	//-----------------------------------------------------
	// Persistent threads mainloop
	//-----------------------------------------------------

	while(1)
	{
		int rayidx;
		{
			int tidx = threadIdx.x;
			int widx = threadIdx.y;
			int popc = 32;
			int eidx = tidx;

			volatile int& localPoolRayCount = rayCountArray[widx];
			volatile int& localPoolNextRay = nextRayArray[widx];

			// Assign rays from global to local pool
			if(eidx==0 && localPoolRayCount<=0)	// local pool empty, eidx 0 fetches
			{
				localPoolNextRay  = atomicAdd(&g_warpCounter, LOAD_BALANCER_BATCH_SIZE*32);
				localPoolRayCount = LOAD_BALANCER_BATCH_SIZE*32;
			}

			// Assign rays from local pool
			rayidx = localPoolNextRay + eidx;
			if(rayidx >= c_rayCount)
				return;

			if(eidx==0)
			{
				localPoolNextRay  += popc;
				localPoolRayCount -= popc;
			}
		}

		//-----------------------------------------------------
		// Fetch and init ray
		//-----------------------------------------------------

		float4 o = FETCH_GLOBAL(rays, rayidx);
		float4 d = FETCH_GLOBAL(rays, rayidx + RAY_BUFFER_SIZE);
		float origx = o.x;
		float origy = o.y;
		float origz = o.z;
		float ooeps = 1/powf(2,-80);
		aux->idir_x  = d.x==0.f ? ooeps : 1/d.x;	// avoid div by 0
		aux->idir_y  = d.y==0.f ? ooeps : 1/d.y;
		aux->idir_z  = d.z==0.f ? ooeps : 1/d.z;
		RayResult* rayResult = &results[__float_as_int(d.w)];
		rayResult->tidx = -1;
		traversalStack[RAY_RESULT_POS] = (int)rayResult;

		//-----------------------------------------------------
		// Traversal init
		//-----------------------------------------------------

		traversalStack[0] = ENTRYPOINT_SENTINEL;	// init traversal stack
		int traversalStackPtr = 0;
		int nodeAddr = root;
		int triAddr  = 0;							// for supporting partially processed leaf nodes
		int triAddr2 = 0;
		float tTriangle= 1.f;

		//------------------------------------------------------
		// TRAVERSE!
		//------------------------------------------------------

		while(nodeAddr!=ENTRYPOINT_SENTINEL || triAddr<triAddr2)
		{
			//---------------------------------------------
			// Inner node
			//---------------------------------------------

			for(int i=LOOP_I-1;i>=0 && nodeAddr>=0 && nodeAddr!=ENTRYPOINT_SENTINEL; i--)
			{
				// Fetch 2 child nodes (boxes + header)
				N_AOS(float4 n0xy = FETCH_TEXTURE(innernodesX, nodeAddr*4+0);)	// node0: x0,x1,y0,y1
				N_AOS(float4 nz   = FETCH_TEXTURE(innernodesX, nodeAddr*4+2);)	// node0: z0,z1, node1: z0,z1
				N_AOS(float4 n1xy = FETCH_TEXTURE(innernodesX, nodeAddr*4+1);)	// node1: x0,x1,y0,y1
				N_AOS(float4 cnodes=FETCH_TEXTURE(innernodesX, nodeAddr*4+3);)
		
				N_SOA(float4 n0xy = FETCH_TEXTURE(innernodesX, nodeAddr);)		// node0: x0,x1,y0,y1
				N_SOA(float4 nz   = FETCH_TEXTURE(innernodesZ, nodeAddr);)		// node0: z0,z1, node1: z0,z1
				N_SOA(float4 n1xy = FETCH_TEXTURE(innernodesY, nodeAddr);)		// node1: x0,x1,y0,y1
				N_SOA(float4 cnodes=FETCH_TEXTURE(nodeheaders, nodeAddr);)

				// Perform 2 ray-box tests
				float ood_x  = origx * aux->idir_x;
				float ood_y  = origy * aux->idir_y;
				float ood_z  = origz * aux->idir_z;
				float x0Child0 = n0xy.x * aux->idir_x - ood_x;
				float x1Child0 = n0xy.y * aux->idir_x - ood_x;
				float y0Child0 = n0xy.z * aux->idir_y - ood_y;
				float y1Child0 = n0xy.w * aux->idir_y - ood_y;
				float z0Child0 = nz.x   * aux->idir_z - ood_z;
				float z1Child0 = nz.y   * aux->idir_z - ood_z;
				float z0Child1 = nz.z   * aux->idir_z - ood_z;
				float z1Child1 = nz.w   * aux->idir_z - ood_z;
				float tminChild0 = fmaxf(0.f,      fminf(x0Child0, x1Child0));
				float tmaxChild0 = fminf(tTriangle,fmaxf(x0Child0, x1Child0));
				float tminChild1 = fmaxf(0.f,      fminf(z0Child1, z1Child1));
				float tmaxChild1 = fminf(tTriangle,fmaxf(z0Child1, z1Child1));
				tminChild0 = fmaxf(tminChild0, fminf(y0Child0, y1Child0));
				tmaxChild0 = fminf(tmaxChild0, fmaxf(y0Child0, y1Child0));
				tminChild0 = fmaxf(tminChild0, fminf(z0Child0, z1Child0));
				tmaxChild0 = fminf(tmaxChild0, fmaxf(z0Child0, z1Child0));
				float x0Child1 = n1xy.x * aux->idir_x - ood_x;
				float x1Child1 = n1xy.y * aux->idir_x - ood_x;
				float y0Child1 = n1xy.z * aux->idir_y - ood_y;
				float y1Child1 = n1xy.w * aux->idir_y - ood_y;
				tminChild1 = fmaxf(tminChild1, fminf(y0Child1, y1Child1));
				tmaxChild1 = fminf(tmaxChild1, fmaxf(y0Child1, y1Child1));
				tminChild1 = fmaxf(tminChild1, fminf(x0Child1, x1Child1));
				tmaxChild1 = fminf(tmaxChild1, fmaxf(x0Child1, x1Child1));

				int traverseChild0 = (tmaxChild0 >= tminChild0);
				int traverseChild1 = (tmaxChild1 >= tminChild1);

				//------------------------------------------------------
				// Traversal decision
				// - this organization compiled to the fastest code
				//------------------------------------------------------

				nodeAddr		   = __float_as_int(cnodes.x);		// stored as int
				int nodeAddrChild1 = __float_as_int(cnodes.y);		// stored as int

				if(traverseChild0 != traverseChild1)
				{
					if(traverseChild1)
						nodeAddr = nodeAddrChild1;
				}
				else
				{
					if(!traverseChild0)	// Neither
					{
						nodeAddr = traversalStack[traversalStackPtr];
						--traversalStackPtr;
					}
					else				// Both
					{
						if(tminChild1 < tminChild0)					// Ensure Child0 is near, Child1 is far
							swap(nodeAddr,nodeAddrChild1);			// Happens if child1 intersected AND closer
						++traversalStackPtr;								// a separate statement thanks to 2.1 compiler shortcoming
						traversalStack[traversalStackPtr] = nodeAddrChild1;	// push far
					}
				}
			}

			//------------------------------------------------------
			// Enter leaf
			//------------------------------------------------------

			if(triAddr >= triAddr2)									// Must be done with the previous leaf
			if(nodeAddr<0)
			{
				// Fetch node header
				N_AOS(float4 leaf=FETCH_TEXTURE(innernodesX, (-nodeAddr-1)*4+3);)
				N_SOA(float4 leaf=FETCH_TEXTURE(nodeheaders, (-nodeAddr-1));)
				triAddr  = __float_as_int(leaf.x);					// stored as int
				triAddr2 = __float_as_int(leaf.y);					// stored as int

				// POP stack
				nodeAddr = traversalStack[traversalStackPtr];
				--traversalStackPtr;
			}

			//---------------------------------------------------------
			// Ray-triangle intersections (Woop's unit triangle test)
			//---------------------------------------------------------

			for(int i=LOOP_T-1; i>=0 && triAddr<triAddr2; triAddr++,i--)
			{
				T_AOS(float4 v00 = FETCH_GLOBAL(trianglesA, triAddr*4+0);)
				T_AOS(float4 v11 = FETCH_GLOBAL(trianglesA, triAddr*4+1);)
				T_SOA(float4 v00 = FETCH_GLOBAL(trianglesA, triAddr);)
				T_SOA(float4 v11 = FETCH_GLOBAL(trianglesB, triAddr);)

				float dirx  = 1/aux->idir_x;
				float diry  = 1/aux->idir_y;
				float dirz  = 1/aux->idir_z;

				float Oz = v00.w - origx*v00.x - origy*v00.y - origz*v00.z;
				float invDz = 1.f / (dirx*v00.x + diry*v00.y + dirz*v00.z);
				float t = Oz * invDz;
				if (t>0.f && t < tTriangle)
				{
					float Ox = v11.w + origx*v11.x + origy*v11.y + origz*v11.z;
					float Dx = dirx*v11.x + diry*v11.y + dirz*v11.z;
					float u = Ox + t*Dx;
					if (u >= 0)
					{
						T_AOS(float4 v22 = FETCH_GLOBAL(trianglesA, triAddr*4+2);)
						T_SOA(float4 v22 = FETCH_GLOBAL(trianglesC, triAddr);)
						float Oy = v22.w + origx*v22.x + origy*v22.y + origz*v22.z;
						float Dy = dirx*v22.x + diry*v22.y + dirz*v22.z;
						float v = Oy + t*Dy;
						if (v >= 0 && (u+v) <= 1.f)
						{
							tTriangle = t;

							RayResult* rayResult = (RayResult*)traversalStack[RAY_RESULT_POS];
							rayResult->tidx = __float_as_int(FETCH_GLOBAL(trianglesI, triAddr));
							rayResult->t	= t;
							if(shadowRay)
							{
								nodeAddr = ENTRYPOINT_SENTINEL;
								triAddr  = triAddr2;	// traversal while
								break;
							}
						}
					}
				}
			}
		}

		// Ray traced, fetch a new one
	}
}
